# $Id: Expression.pm,v 1.7 2007/02/20 17:39:38 faeder Exp $

package Expression;

use Class::Struct;
use FindBin;
use lib $FindBin::Bin;
use Param;
use ParamList;

struct Expression => {
  Type=> '$',
  Arglist=>'@',
  Err=>'$',
};

%functions=(
	    "exp"=>{FPTR=>sub{exp($_[0])},NARGS=>1},
	    "cos"=>{FPTR=>sub{cos($_[0])},NARGS=>1},
	    "sin"=>{FPTR=>sub{sin($_[0])},NARGS=>1},
	    "log"=>{FPTR=>sub{log($_[0])},NARGS=>1},
	    "abs"=>{FPTR=>sub{abs($_[0])},NARGS=>1},
	    "int"=>{FPTR=>sub{int($_[0])},NARGS=>1},
	    "sqrt"=>{FPTR=>sub{sqrt($_[0])},NARGS=>1},
	   );

my $MAX_LEVEL=100; # Prevent infinite loop due to dependency loops

{
  my $string_sav;
  my %variables;

  sub readString{
    my $expr= shift;
    my $sptr= shift;
    my $plist = (@_) ? shift : "";
    my $end_chars= (@_) ? shift : "";
    my $level = (@_) ? shift : 0;
    my $err="";
    my $ops_bi='\*\*|[+\-*/^]';
    my $ops_un='[+-]';

    if (!$level){
      $string_sav= $$sptr;
      %variables=();
    }
    
    # parse string into form expr op expr op ... 
    # a+b*(c+d)
    # -5.0e+3/4
    my $last_read="";
    my $expr_new="";
    my @list=();
    my $expect_op=0;
    my $assign_var_name="";
    while($$sptr ne ""){
      #print "string=$$sptr\n";
      if ($expect_op==1){
	# OPERATOR
	if ($$sptr=~s/^\s*($ops_bi)//){
	  my $express= Expression->new();
	  $express->Type($1);
	  push @list, $express;
	  $expect_op=0;
	  next;
	} 
	# Assignment using '='.  Valid syntax is VAR = EXPRSSION
	elsif ($$sptr=~ s/^\s*=//){
	  # Check that only preceding argument is variable
	  my $var= $list[$#list];
	  my $vname= $var->Arglist->[0];
	  my $var_count_start=$variables{$vname};
	  if ($#list>0){
	    return("Invalid assignment syntax (VAR = EXPRESSION) in $string_sav at $$sptr");
	  }
	  #if ($level>0){
	  #  return("Nested assignment of variables is not allowed in $string_sav at $$sptr");
	  #}
	  if ($var->Type ne 'VAR'){
	    return("Attempted assignment to non-variable type in $string_sav at $$sptr.");
	  }
	  # Read remainder as expression
	  my $rhs= Expression->new();
	  $err=$rhs->readString($sptr, $plist, $end_chars, $level+1);
	  $err && return($err);
	  # Perform assignment of expression to variable
	  # Evaluate rhs if lhs VAR occurs on rhs
	  if ($variables{$vname}>$var_count_start){
	    $plist->set($vname, $rhs->evaluate($plist));
	  } 
	  # otherwise, set variable to rhs.
	  else {
	    $plist->set($vname, $rhs);
	  }
	  #print "Evaluates to ", $expr->evaluate($plist),"\n";
	  #print "Prints    to ", $expr->toString($plist),"\n";
	  last;
	}
	# Look for end characters
	elsif ($end_chars && ($$sptr=~ /^\s*${end_chars}/)){
	  last;
	} else {
	  $$sptr=~ s/^\s*//;
	  #last unless ($$sptr);
	  last;
	  return("Expecting operator in expression $string_sav at $$sptr");
	}
      }

      # Chop leading whitespace
      $$sptr=~s/^\s*//;

      # NUMBER
      if (my $express=getNumber($sptr)){
	#printf "Read NUM %s\n", $express->evaluate();
	push @list, $express;
	$expect_op=1;
	next;
      }
      
      # FUNCTION
      if ($$sptr=~s/^($ops_un)?\s*([A-Za-z0-9_]+)\s*\(//){
	if (my $op= $1){
	  # (optional) UNARY OP at start of expression, as in -a + b, or -a^2
	  my $express_u= Expression->new();
	  $express_u->Type($1);
	  push @list, $express_u;
	}
	my $name= $2;
	my @fargs=();
	if (defined($functions{$name})){
	  while(1){
	    my $express= Expression->new();
	    $err=$express->readString($sptr, $plist, ',\)', $level+1);
	    $err && return($err);
	    push @fargs, $express;
	    if ($$sptr=~ s/^\)//){
	      last;
	    }
	    elsif ($$sptr=~ s/^,//){
	      next;
	    }
	  }
          if ($functions{$name}->{NARGS}!=scalar(@fargs)){
            return("Incorrect number of arguments to function $name");
          }
	} else {
	  return("Undefined function $name");
	}
	my $express= Expression->new();
	$express->Type('FUN');
	$express->Arglist([$name,@fargs]);
	push @list, $express;
	$expect_op=1;
	next;
      }
      # VARIABLE 
      elsif ($$sptr=~s/^($ops_un)?\s*([A-Za-z0-9_]+)//){
	if (my $op= $1){
	  # (optional) UNARY OP at start of expression, as in -a + b, or -a^2
	  my $express_u= Expression->new();
	  $express_u->Type($1);
	  push @list, $express_u;
	}
	my $name= $2;
	# Validate against ParamList, if present
	if ($plist){
	  # Create and set variable if next token is '='
	  # otherwise create referenced variable but leave its Expr unset
	  if (!($$sptr=~ /^\s*=/)){
	    my ($param, $err)= $plist->lookup($name);
	    if (!$param){
              #return("Can't reference undefined parameter $name");
	      $plist->set($name);
	    }
	  }
	} else {
	  return("No parameter list provided");
	}
	my $express= Expression->new();
	$express->Type('VAR');
	$express->Arglist([$name]);
	++$variables{$name};
	push @list, $express;
	$expect_op=1;
	next;
      }
      # Get expression enclosed in parenthesis
      elsif($$sptr=~s/^($ops_un)?\s*\(//){
	if (my $op= $1){
	  # (optional) UNARY OP at start of expression, as in -a + b, or -a^2
	  my $express_u= Expression->new();
	  $express_u->Type($1);
	  push @list, $express_u;
	}
	my $express= Expression->new();
	$err=$express->readString($sptr, $plist, '\)', $level+1);
	$err && return($err);
	if (!($$sptr=~ s/^\)//)){
	  return("Missing end parenthesis in $string_sav at $$sptr");
	}
	#printf "express=%s %s\n", $express->toString($plist), $$sptr;
	push @list, $express;
	$expect_op=1;
	next;
      }
      # ERROR
      else {
	return("Expecting operator argument in $string_sav at $$sptr");
      }      
    }
    
    # Transform list into expression preserving operator precedence
    $expr->copy(arrayToExpression(@list));

    return($err);
  }
}

{
  sub depends{
    my $expr=shift;
    my $plist= shift;
    my $varname=shift;
    my $level= (@_) ? shift : 0;
    my $dep= (@_) ? shift : {};

    my $retval="";
    my $err="";

    my $type= $expr->Type;
    if ($type eq 'NUM'){
    } 
    elsif($type eq 'VAR'){
      #printf "type=$type %s\n", $expr->toString($plist);
      my $vname= $expr->Arglist->[0];
      #print "$varname $vname\n";
      if ($$dep{$vname}){
	$err= sprintf "Cycle in parameter $vname looking for dep in %s", $varname;
	print "$err\n";
	$retval=$vname;
      } else {
	#++$$dep{$vname};
	if($varname eq $vname){
	  $retval=$vname;
	}
	elsif ($plist){
	  my $param;
	  ($param, $err)= $plist->lookup($vname);
	  my %newdep= %{$dep};
	  $newdep{$nvame}= 1;
	  (my $ret, $err)= ($param->Expr->depends($plist,$varname,$level+1,\%newdep));
	  if ($ret){
	    $retval= $param->Name.'->'.$ret;
	  }
	}
      }
    }
    else{
      my @arglist= @{$expr->Arglist};
      # Skip function name if this is a function
      if ($type eq 'FUN'){shift(@arglist);}
      for my $e (@arglist){
	($retval, $err)= ($e->depends($plist,$varname,$level+1,$dep));
	last if $retval;
      }
    }    
    #print "level=$level $retval $err\n";
    return($retval,$err);
  }
}

sub copy{
  my $edest=shift;
  my $esource=shift;

  $edest->Type($esource->Type);
  $edest->Arglist([@{$esource->Arglist}]);
  return ($edest);
}

sub evaluate{
  my $expr= shift;
  my $plist = (@_) ? shift : "";
  my $level= shift;

  ($level > $MAX_LEVEL) && die "Max recursion depth $MAX_LEVEL exceeded.";

  my $val;
  my $alist= $expr->Arglist;
  my $type= $expr->Type;
  if ($type eq 'NUM'){
    $val= $$alist[0];
  } 
  elsif($type eq 'VAR'){
    if (!$plist){
      $expr->Err("Variable is referenced but not ParamList is provided.");
      return("");
    }
    
    $val= $plist->evaluate($$alist[0], $level+1);
    if ($err){
      $expr->Err($err);
      return("");
    }
  }
  elsif($type eq 'FUN'){
    my @args= @{$expr->Arglist};
    my $name= shift(@args);
    my $f= $functions{$name}->{FPTR};
    my @fargs=();
    for my $arg(@args){
      push @fargs, $arg->evaluate($plist);
    }
    $val= $f->(@fargs);
  }
  elsif($type eq '+'){
    for my $e (@$alist){
      $x = $e->evaluate($plist,$level+1);
      $val+= $x;
    }
  }
  elsif($type eq '*'){
    $val=1;
    for my $e (@$alist){
      $val*= $e->evaluate($plist,$level+1);
    }
  }
  elsif($type eq '-'){
    if ($#$alist==0){
      $val= - $$alist[0]->evaluate($plist,$level+1);
    } else {
      $val= $$alist[0]->evaluate($plist,$level+1) - $$alist[1]->evaluate($plist,$level+1);
    }
  }
  elsif($type eq '/'){
    if ($#$alist==0){
      $val= 1/$$alist[0]->evaluate($plist,$level+1);
    } else {
      $val= $$alist[0]->evaluate($plist,$level+1) / $$alist[1]->evaluate($plist,$level+1);
    }
  }
  elsif($type eq '^'){
    $val= ($$alist[0]->evaluate($plist,$level+1))**($$alist[1]->evaluate($plist,$level+1));
  }
  elsif($type eq '**'){
    $val= ($$alist[0]->evaluate($plist,$level+1))**($$alist[1]->evaluate($plist,$level+1));
  }
  else{
    $expr->Err("Unrecognized Expression type ".$type);
    return("");
  }
  
  return($val);
}

sub toString{
  my $expr= shift;
  my $plist = (@_) ? shift : "";
  my $level= (@_) ? shift : 0;
  
  ($level > $MAX_LEVEL) && die "Max recursion depth $MAX_LEVEL exceeded.";

  my $string="";
  my $type= $expr->Type;
  if ($type eq 'NUM'){
    $string= $expr->Arglist->[0];
    #print "NUM=$string\n";
  } 
  elsif($type eq 'VAR'){
    $string= $expr->Arglist->[0];
    #$string= $expr->evaluate($plist);
    #print "VAR=$string\n";
  }
  elsif($type eq 'FUN'){
    my @sarr=();
    for my $i (1..$#{$expr->Arglist}){
      push @sarr, $expr->Arglist->[$i]->toString($plist,$level+1);
    }
    $string= $expr->Arglist->[0].'('.join(',',@sarr).')';
  }
  else{
    my @sarr=();
    for my $e (@{$expr->Arglist}){
      push @sarr, $e->toString($plist,$level+1);
    }
    if ($#sarr>0){
      $string= join($type,@sarr);
    } else {
      $string= $type.$sarr[0];
    }
    # enclose in brackets if not at top level
#    print "level=$level\n";
    if ($level){
      $string= '('.$string.')';
    }
    #printf "%s=$string\n", $expr->Type;
  }
  
  return($string);
}


{
  my %ophash= (
	       '+'=>'plus',
	       '-'=>'minus',
	       '*'=>'times',
	       '/'=>'divide',
	       '**'=>'power',
	       '^'=>'power',
	      );

  sub toMathMLString{
    my $expr= shift;
    my $plist = (@_) ? shift : "";
    my $indent= (@_) ? shift : "";
    my $level= (@_) ? shift : 0;
 
    ($level > $MAX_LEVEL) && die "Max recursion depth $MAX_LEVEL exceeded.";

    my $string="";
    my $indentp=$indent;
    if ($level==0){
      $string.= $indent."<math xmlns=\"http://www.w3.org/1998/Math/MathML\">\n";
      $indentp.= "  ";
    }      

    my $type= $expr->Type;
    if ($type eq 'NUM'){
      $string.= sprintf "%s<cn> %s </cn>\n", $indentp, $expr->Arglist->[0];
    } 
    elsif($type eq 'VAR'){
      $string.= sprintf "%s<ci> %s </ci>\n", $indentp, $expr->Arglist->[0];
    }
    elsif($type eq 'FUN'){
      $string.= $indentp."<apply>\n";
      my $indentpp=$indentp."  ";
      my @arglist= @{$expr->Arglist};
      $string.= sprintf "%s<%s/>\n", $indentpp, shift(@arglist);
      for my $e (@arglist){
	$string.= $e->toMathMLString($plist,$indentpp,$level+1);
      }
      $string.= $indentp."</apply>\n";
    }
    else{
      $string.= $indentp."<apply>\n";
      my $indentpp=$indentp."  ";
      $string.= sprintf "%s<%s/>\n", $indentpp, $ophash{$expr->Type};
      for my $e (@{$expr->Arglist}){
	$string.= $e->toMathMLString($plist,$indentpp,$level+1);
      }
      $string.= $indentp."</apply>\n";
    }
    
    if ($level==0){
      $string.= $indent."</math>\n";
    }
    return($string);
  }
}


# Convert an array of type EXPR OP EXPR OP ... to a single Expression.
sub arrayToExpression{
  my @earr= @_;

  # list of optypes in order of precedence
  my @operators= ('\*\*|\^','[*/]','[+-]');
  my $optype= shift @operators;
  while($optype){
    my $i=0;
    # Consolidate EXPR OP EXPR into EXPR
    while ($i<$#earr){
      my $expr= $earr[$i];
      if ( $expr->Type=~ /$optype/ && !(@{$expr->Arglist})){
	if ($i>0){
	  $expr->Arglist->[0]= $earr[$i-1];
	  $expr->Arglist->[1]= $earr[$i+1];
	  splice @earr, $i-1, 3, $expr;
	  next;
	} else {
	  # Handle leading unary op, as in -a + b
	  $expr->Arglist->[0]= $earr[$i+1];
	  splice @earr, $i, 2, $expr;
	  ++$i;
	  next;
	}
      }
      ++$i;
    }
    #print "expression after $optype= ";
    #foreach $expr (@earr){
    # printf " %s", $expr->toString();
    #}
    #print "\n";
    # Finished with current optype
    $optype= shift @operators;
  }  

  #printf "final expression= %s\n", $earr[0]->toString();

  return($earr[0]);
}

sub getNumber{
  my $string=shift;
  my $number="";

  # Decimal part
  if ($$string=~ s/^([+-]?\d+[.]?\d*)//){
    $number=$1;
  } 
  elsif ($$string=~ s/^([+-]?[.]\d+)//){
    $number=$1;
  } else {
    return("");
  }

  # Exponent part
  if ($$string=~ s/^([DEFGdefg][+-]?\d+)//){
    $number.=$1;
  } 
  elsif ($$string=~ /^[A-Za-z_]/){
    # String is non a number; restore value of string
    $$string= $number.$$string;
    #print "gothere: $$string\n";
    return("");
  }
  my $express= Expression->new();
  $express->Type('NUM');
  $express->Arglist([$number]);
  return($express);
}

# Returns name of VAR if expression of VAR or 
# return name of new VAR containing expression.
sub getName{
  my $expr=shift;
  my $plist= shift;
  my $basename= (@_) ? shift : "k";
  my $name;
  if ($expr->Type eq 'VAR'){
    $name= $expr->Arglist->[0];
    #printf "Found existing parameter %s\n", $name;
  } else {
    # Find unused name 
    my $index=1;
    while(1){
      my ($param, $err)= $plist->lookup($basename.$index);
      last unless $param;
      ++$index;
    }
    $name= $basename.$index;
    $plist->set($name,$expr);
    #printf "Creating new parameter %s\n", $name;
  }

  return($name);
}


1;
