package Map;

use Class::Struct;
use FindBin;
use lib $FindBin::Bin;
use BNGUtils;
use SpeciesGraph;

struct Map => {
  Source=> 'SpeciesGraph',	       
  Target=> 'SpeciesGraph',	       
  MapF=>   '%',	       # Forward mapping of Source onto Target
  MapR=>   '%',	       # Reverse mapping of Target onto Source
};

sub toString{
  my $map= shift;

  my $string="";
  
  *mapf= $map->MapF;
  $sg= $map->Source;
  *molecules= $sg->Molecules;

  for my $im (0..$#molecules){
    $im2= $mapf{$im};
    #$string.= sprintf "Mol %d -> Mol %d\n", $im, $im2;
    $string.= sprintf " %d->%d(", $im+1, $im2+1;
    next if ($im2<0);
    *components= $molecules[$im]->Components;
    for my $ic (0..$#components){
      my ($im2, $ic2)= split('\.', $mapf{"$im.$ic"});
      #$string.= sprintf "  Comp %d -> Comp %d\n", $ic, $ic2;
      $string.= sprintf " %d->%d", $ic+1, $ic2+1;
    }
    $string.=")";
  }
  return($string);
}

sub mapF{
  my $map=shift;
  my $p1= shift;
  my $offset= (@_) ? shift : 0;

  *mapf= $map->MapF;
  my $p2= $mapf{$p1};
  if ($offset){
    my ($im2, $ic2)= split('\.', $p2);
    $im2+= $offset;
    if ($ic2 ne ""){
      $p2= "$im2.$ic2";
    } else {
      $p2= "$im2";
    }
  }
  
  return $p2;
}

# Transform self-map according to map given as argument
sub mapAuto{
  my $map= shift;
  my $tmap= shift;
  
  if ($map->Source != $map->Target){
    return("mapAuto called on non-self map");
  }
  
  if ($map->Source != $tmap->Source){
    return("mapAuto: Argument map not compatible with map");
  }

  # Change map Source and Target
  $map->Source($tmap->Target);
  $map->Target($tmap->Target);

  *mapf= $map->MapF;
  *tmapf= $tmap->MapF;

  # Set up identity map
  my %mapnew=();
  *molecules= $tmap->Target->Molecules;
  for my $im (0..$#molecules){
    $mapnew{$im}=$im;
    *components= $molecules[$im]->Components;
    for my $ic (0..$#components){
      $mapnew{"$im.$ic"}= {"$im.$ic"};
    }  
  }
  
  # Transform self-map
  for my $p1 (keys %mapf){
    my $p2= $mapf{$p1};
    my $tp1= $tmapf{$p1};
    my $tp2= $tmapf{$p2};
    next if ($tp1<0);
    $mapnew{$tp1}=$tp2;
  }
  
  # Replace MapF (NOTE: MapR is not changed! This is  because it should not be
  # used by self-map, which is generated by calling subGraphIsomorphism))
  $map->MapF({%mapnew});

  return("");
}

1;
