# $Id: RxnRule.pm,v 1.20 2007/07/06 06:09:48 faeder Exp $

package RxnRule;

use Class::Struct;
use FindBin;
use lib $FindBin::Bin;
use BNGUtils;
use SpeciesGraph;
use RateLaw;
use SpeciesList;
use MoleculeTypesList;
use RxnList;
#use strict;

struct RxnRule => {
  Name=>      '$',		   
  Reactants=> '@',  # Arrays of SpeciesGraphs
  Products=> '@',
  Rmatches=>  '@',   # Array of Maps onto SpeciesGraphs selected by Reactants SpeciesGraphs
  Rinclude=>  '@',   # Array of patterns to use as context for selecting reactants
  Pinclude=>  '@',   # Array of patterns to use as context for selecting products
  Rexclude=>  '@',   # Array of patterns to exclude from reactant matches		   
  Pexclude=>  '@',   # Array of patterns to exclude from products
  Map=>       '%',
  RateLaw=> 'RateLaw',   # Array of rate laws, 1 if unidirectional, 2 if reversible
  Priority=> '$',	   
  MultScale=> '$', #Factor by which to scale multiplicity to account for symmetry in rule		 
# The operations performed by the rule
  EdgeAdd=> '@',
  EdgeDel=> '@',
  MolAdd=>   '@',
  MolDel=>   '@',
  CompStateChange=> '@',
  ChangeCompartment=> '@',
  DeleteMolecules=> '$' # If True, deleting a molecule in a species will remove only that molecule.
                       # Default is unset, and behavior of deletion is determined automatically.
};

sub newRxnRule{
  my $string= shift;
  my $plist= shift;
  my $clist= shift;
  my $mtlist= (@_) ? shift : "";
  my $err,$sep,$reversible,$DeleteMolecules;
  my $name="";
  
  $DeleteMolecules=0;
  $name= "";

  if ($string=~ s/^\s*([^:].*)[:]\s*//){
    # Check if first token is a name for the reaction (first occurence of a ':')
    $name= $1;
    #print "(user) name=$name\n";
  }
  # Check if token is an index
  elsif ($string=~ s/^\s*(\d+)\s+//){
    $name=$1;
    #print "(index) name=$name\n";
  }
  
  # Remove leading whitespace
  $string=~ s/^\s*//;

  if ($string=~ s/CreateSpecies\s*\(\s*//i){
    # Syntax CreateSpecies(SpeciesGraph, RateLaw)

    # Read SpeciesGraph- checking that it represents a valid species
    my $g= SpeciesGraph->new();
    $err= $g->readString(\$string,$clist,1,'^\s*[,]\s*',$mtlist);
    if ($err){return([],$err)};

    if (!($string=~ s/^\s*[,]\s*//)){
      return([],"Arguments must be separated by , in CreateSpecies: $string");
    }

    # Read RateLaw
    ($rl,$err)= RateLaw::newRateLaw(\$string,$plist);
    if ($err){ return([],$err);}

    if (!($string=~ s/^\s*\)\s*//)){
      return([],"No end parenthesis in CreateSpecies: $string");
    }
    
    if ($string){
      return([],"Syntax error in CreateSpecies: $string");
    }

    return([],"CreateSpecies not yet implemented.");

    # Create mapping 
    # Problem here arising from null Reactant list
    if ($err=$rr->findMap($mtlist)){
      return([],$err);
    }

    # Construct RxnRule
    my $rr= RxnRule->new();
    $rr->Name($name);
    $rr->Reactants([]);
    $rr->Products([$g]);
    $rr->RateLaw($rl);

    return([$rr],"");
  }
  elsif ($string=~ s/DeleteSpecies\s*\(\s*//i){
    return([],"DeleteSpecies not yet implemented.");
  }
  elsif ($string=~ s/DeleteMolecules\s*\(\s*//i){
    return([],"DeleteMolecules not yet implemented.");
  }
  # READ REACTION IN STANDARD SYNTAX
  else{
    # read reactant patterns
    my @reac=();
    my %labels=();
    $sep='^\s*[+]\s*|^\s*([<]?-[>])\s*';
    my $g;
    while($string){
      $g= SpeciesGraph->new();
      $err= $g->readString(\$string,$clist,0,$sep,$mtlist);
      if ($err){return([],$err)};
      # Checking labels for uniqueness
      for my $mol (@{$g->Molecules}){
	$label= $mol->Label;
	if ($label ne ""){
	  if ($labels{$label}){
	    return([],"Repeated label $label in reactants of reactant rule");
	  }
	  $labels{$label}="M";
	}
	for my $comp (@{$mol->Components}){
	  $label= $comp->Label;
	  if ($label ne ""){
	    if ($labels{$label}){
	      return([],"Repeated label $label in reactants of reaction rule");
	    }
	    $labels{$label}="C";
	  }
	}
      }
      push @reac, $g;
      if (!($string=~ s/$sep//)){
      return([],"Invalid syntax in reaction rule: $string");
    }
      if ($1){
	$reversible= ($1 eq "<->") ? 1 : 0;
	last;
      }
    }
    
    # read product patterns
    my @prod=();
    $sep='^\s*([+])\s*|^\s+';
    while($string){
      my $g= SpeciesGraph->new();
      $err= $g->readString(\$string,$clist,0,$sep,$mtlist);
      if ($err){return([],$err)};
      $string=~ s/$sep//;
    # Checking labels for correspondence and uniqueness
      for my $mol (@{$g->Molecules}){
	$label= $mol->Label;
	if ($label ne ""){
	  if ($labels{$label} eq "M"){
	  $labels{$label}= "found";
	}
	  elsif ($labels{$label} eq "found"){
	    return([],"Repeated label $label in products of reaction rule $name");
	  } else {
	    return([],"Mis- or un-matched label $label in products of reaction rule $name");
	  }
	}
	for my $comp (@{$mol->Components}){
	  $label= $comp->Label;
	  #print "label=$label\n";
	  if ($label ne ""){
	    if ($labels{$label} eq "C"){
	      $labels{$label}= "found";
	    }
	    elsif ($labels{$label} eq "found"){
	      return([],"Repeated label $label in products of reaction rule $name");
	    } else {
	      return([],"Mis- or un-matched label $label in products of reaction rule $name");
	    }
	  } 
	  
	}
      }
      push @prod, $g;
      last unless $1;
    }
    
    # Any label not found 
    
    # read rate constants 
    my $rl, @rate_laws=();
    ($rl,$err)= RateLaw::newRateLaw(\$string,$plist);
    if ($err){ return([],$err);}
    push @rate_laws, $rl;
    
    # Reversible reaction requires two valid rate laws
    if ($reversible){
      # Remove separator for backward compatibility
      $string=~ s/^\s*[,]\s*//;
      ($rl,$err)= RateLaw::newRateLaw(\$string,$plist);
      if ($err){ return([],$err);}
      push @rate_laws, $rl;
    } else {
      if ($string=~/^\s*[,]/){
	return([],"Unidirection reaction may have only one rate law");
      }
    }

    # Set optional priority of rule
    $priority=0;
    if ($string=~ s/priority\s*=\s*(\S+)//){$priority=$1;}
    
    # Set DeleteMolecules attribute
    $attr="DeleteMolecules";
    if ($string=~ s/$attr//){
      $DeleteMolecules= 1;
    }
    
    # Process exclude_reactants
    my @Rexclude=(())x@reac;
    my @Pexclude=(())x@prod;
    while($string=~ s/exclude_(\w+)(\S+)//){
      my $type= $1;
      my $arg= $2;
      if (!($arg=~ s/^\(//)){
	return([],"Arguments to exclude_$type must be enclosed in parenthesis");
      }
      if (!($arg=~ s/\)$//)){
	return([],"exclude_$type missing close parenthesis");
      }
      
      # First argument is number of reactant or product
      if (!($arg=~ s/^\s*(\d+)\s*[,]//)){
	return([],"exclude_$type first argument must be index of $type");
      }
      my $index=$1-1;
      
      if ($type eq "reactants"){
	my $max_index= $#reac;
	if ($index<0 || $index>$max_index){
	  return([],"exclude_$type index $index out of range");
	}
	my $sep='^,';
	while($arg){
	  $g= SpeciesGraph->new();
	  $err= $g->readString(\$arg,$clist,0,$sep,$mtlist);
	  $arg=~ s/$sep//;
	  if ($err){return([],$err)};
	  push @{$Rexclude[$index]}, $g;
	}
      }
      elsif ($type eq "products"){
	my $max_index= $#prod;
	if ($index<0 || $index>$max_index){
	  return([],"exclude_$type index $index out of range");
	}
	my $sep='^,';
	while($arg){
	  $g= SpeciesGraph->new();
	  $err= $g->readString(\$arg,$clist,0,$sep,$mtlist);
	  $arg=~ s/$sep//;
	  if ($err){return([],$err)};
	  push @{$Pexclude[$index]}, $g;
	}
      }
      else{
	return([],"Unrecognized function exclude_$type");
      }
    }
    
    # Process include_reactants
    my @Rinclude=(())x@reac;
    my @Pinclude=(())x@prod;
    while($string=~ s/include_(\w+)(\S+)//){
      my $type= $1;
      my $arg= $2;
      if (!($arg=~ s/^\(//)){
	return([],"Arguments to include_$type must be enclosed in parenthesis");
      }
      if (!($arg=~ s/\)$//)){
	return([],"include_$type missing close parenthesis");
      }
      
      # First argument is number of reactant or product
      if (!($arg=~ s/^\s*(\d+)\s*[,]//)){
	return([],"include_$type first argument must be index of $type");
      }
      my $index=$1-1;
      
      if ($type eq "reactants"){
	my $max_index= $#reac;
	if ($index<0 || $index>$max_index){
	  return([],"include_$type index $index out of range");
	}
	my $sep='^,';
	while($arg){
	  $g= SpeciesGraph->new();
	  $err= $g->readString(\$arg,$clist,0,$sep,$mtlist);
	  $arg=~ s/$sep//;
	  if ($err){return([],$err)};
	  push @{$Rinclude[$index]}, $g;
	}
      }
      elsif ($type eq "products"){
	my $max_index= $#prod;
	if ($index<0 || $index>$max_index){
	  return([],"include_$type index $index out of range");
	}
	my $sep='^,';
	while($arg){
	  $g= SpeciesGraph->new();
	  $err= $g->readString(\$arg,$clist,0,$sep,$mtlist);
	  $arg=~ s/$sep//;
	  if ($err){return([],$err)};
	  push @{$Pinclude[$index]}, $g;
	}
      }
      else{
	return([],"Unrecognized function include_$type");
      }
    }
    
    if ($string=~/\S+/){
      my $err= sprintf "Syntax error in reaction rule at $string";
      return([],$err);
    }

    my @rrs=();
    my $rr= RxnRule->new();
    ($name ne "") && $rr->Name($name);
    $rr->Reactants([@reac]);
    $rr->Products([@prod]);
    $rr->Priority($priority);
    $rr->RateLaw($rate_laws[0]);
    $rr->Rexclude([@Rexclude]);
    $rr->Pexclude([@Pexclude]);
    $rr->Rinclude([@Rinclude]);
    $rr->Pinclude([@Pinclude]);
    $rr->DeleteMolecules($DeleteMolecules);
    
    # Find mapping of reactants onto products
    if ($err=$rr->findMap($mtlist)){
      return([],$err);
    }
    push @rrs, $rr;
    #print "rrule: ",$rr->toString(),"\n";
    
    if ($reversible){
      $rr= RxnRule->new();
      ($name ne "") && $rr->Name("${name}(reverse)");
      $rr->Reactants([@prod]);
      $rr->Products([@reac]);
      $rr->Priority($priority);
      $rr->RateLaw($rate_laws[1]);
      $rr->Pexclude([@Rexclude]);
      $rr->Rexclude([@Pexclude]);
      $rr->Pinclude([@Rinclude]);
      $rr->Rinclude([@Pinclude]);
      $rr->DeleteMolecules($DeleteMolecules);
      # Find mapping of reactants onto products
      if ($err=$rr->findMap($mtlist)){
	return([],$err);
      }
      push @rrs, $rr;
      #print "rrule: ",$rr->toString(),"\n";
    }
    return [@rrs];
  } # END OF STANDARD REACTION SYNTAX

  # Should not get here
  return([],"Unknown error in RxnRule::readString");
 }

sub listActions{
  my $rr=shift;
  my $out="";

#  my %map_reverse=();
#  for my $s (keys %{$rr->Map}){
#    my $t= $rr->Map->{$s};
#    $map_reverse{$t}= $s;
#  }

  # State changes
  for my $sc (@{$rr->CompStateChange}){
    (my $p, my $sR, my $sP)= @{$sc};
    $out.= "# StateChange($p,$sR->$sP)\n";
#    my $p2= $map_reverse{$p};
#    $out.= "# StateChange($p,$sR->$sP) (reactant)\n";
  }

  # Binding
  for my $ea (@{$rr->EdgeAdd}){
    (my $p1, my $p2)= @{$ea};
    $out.= "# Bind($p1,$p2)\n";
#    $p1= $map_reverse{$p1};
#    $p2= $map_reverse{$p2};
#    $out.= "# Bind($p1,$p2) (reactants)\n";
  }

  # Dissociation
  for my $ed (@{$rr->EdgeDel}){
    (my $p1, my $p2)= @{$ed};
    $out.= "# Unbind($p1,$p2)\n";
  }

  # Molecule Deletion
  for my $molref (@{$rr->MolDel}){
    my ($p,$m)= split('\.',$molRef);
    if (defined($m)){
      $out.= "# Delete Molecule ".$rr->Reactants->[$p]->Molecules->[$m]->toString()."\n";
    } else {
      $out.= "# Delete Pattern ".$rr->Reactants->[$p]->toString()."\n";
    }
  }

  # Species Creation
  for my $molAddRef (@{$rr->MolAdd}){
    my @compInhRefs= @$molAddRef;
    my $molRef= shift(@compInhRefs);
    my ($p,$m)= split('\.',$molRef);
    my $newMol= $rr->Products->[$p]->Molecules->[$m];
    $out.= "# Add Species ".$newMol->toString()."\n";
  }

  return($out);
}

sub toString{
  my $rr= shift;
  my $rr_rev= (@_)?shift:0;
  my $string="";

  # Reaction name
  # Name is an index unless it contains non-digit character
  if ($rr->Name=~/\D/){
    $string.= $rr->Name.": \\\n  ";
  } else {
    $string.= sprintf "%5d ",$rr->Name;
  }
  my $i=0;
  for my $g (@{$rr->Reactants}){
    if ($i!=0){ $string.= " + ";}
    $string.= $g->toString();
    ++$i;
  }
  $string.= ($rr_rev) ? " <-> " : " -> ";

  $i=0;
  for $g (@{$rr->Products}){
    if ($i!=0){ $string.= " + ";}
    $string.= $g->toString();
    ++$i;
  }

  $string.= " ".$rr->RateLaw->toString();
  if ($rr_rev){
    $string.= ", ".$rr_rev->RateLaw->toString();
  }

  if ($rr->Priority !=0){
    $string.= sprintf " priority=%d", $rr->Priority;
  }

  for my $i (0..$#{$rr->Rexclude}){
    next unless (@{$rr->Rexclude->[$i]});
    $string.= " exclude_reactants(";
    $string.= $i+1;
    for my $g(@{$rr->Rexclude->[$i]}){
      $string.= ','.$g->toString();
    }
    $string.= ")";
  }

  for my $i (0..$#{$rr->Pexclude}){
    next unless (@{$rr->Pexclude->[$i]});
    $string.= " exclude_products(";
    $string.= $i+1;
    for my $g(@{$rr->Pexclude->[$i]}){
      $string.= ','.$g->toString();
    }
    $string.= ")";
  }

  for my $i (0..$#{$rr->Rinclude}){
    next unless (@{$rr->Rinclude->[$i]});
    $string.= " include_reactants(";
    $string.= $i+1;
    for my $g(@{$rr->Rinclude->[$i]}){
      $string.= ','.$g->toString();
    }
    $string.= ")";
  }

  for my $i (0..$#{$rr->Pinclude}){
    next unless (@{$rr->Pinclude->[$i]});
    $string.= " include_products(";
    $string.= $i+1;
    for my $g(@{$rr->Pinclude->[$i]}){
      $string.= ','.$g->toString();
    }
    $string.= ")";
  }

  if ($rr->DeleteMolecules){
    $string.= " DeleteMolecules";
  }

  return $string;
}

sub toXML{
  my $rr= shift;
  my $indent= shift;
  my $index=shift;

  my $string=$indent."<ReactionRule";

  # Attributes
  # id
  my $id= "RR".$index;
  $string.= " id=\"".$id."\"";
  # name
  if ($rr->Name){
    $string.= " name=\"".$rr->Name."\"";
  }
  # priority
  if ($rr->Priority!=0){
    $string.= " priority=\"".$rr->Priority."\"";
  }

  # Objects contained
  my $ostring="";
  my $indent2= "  ".$indent;
  my $indent3= "  ".$indent2;
  
  # Reactants
  $ostring.=$indent2."<ListOfReactantPatterns>\n";
  my $index=1;
  for my $patt (@{$rr->Reactants}){
    my $pid= $id."_RP".$index;
    $ostring.= $patt->toXML($indent3,"ReactantPattern",$pid,"");
    ++$index;
  }
  $ostring.=$indent2."</ListOfReactantPatterns>\n";

  # Products
  $ostring.=$indent2."<ListOfProductPatterns>\n";
  my $index=1;
  for my $patt (@{$rr->Products}){
    my $pid= $id."_PP".$index;
    $ostring.= $patt->toXML($indent3,"ProductPattern",$pid,"");
    ++$index;
  }
  $ostring.=$indent2."</ListOfProductPatterns>\n";

  # RateLaw
  $ostring.= $rr->RateLaw->toXML($indent2,$id."_RateLaw");

  # Map
  $ostring.=$indent2."<Map>\n";
  my $index=1;
  for my $source (sort keys %{$rr->Map}){
    $ostring.= $indent3."<MapItem";
    my $target=$rr->Map->{$source};
    $ostring.= " sourceID=\"".pointer_to_ID($id."_R",$source)."\"";
    $target=pointer_to_ID($id."_P",$target);
    if ($target != "Null"){
      $ostring.= " targetID=\"".$target."\"";
    }
    $ostring.= "/>\n";
    ++$index;
  }
  $ostring.=$indent2."</Map>\n";
  
  # Operations
  $ostring.=$indent2."<ListOfOperations>\n";
  my $index=1;
  # construct map fro product molecules and components onto reactants
  my %map_reverse=();
  for my $s (keys %{$rr->Map}){
    my $t= $rr->Map->{$s};
    $map_reverse{$t}= $s;
  }

  # State changes
  for my $sc (@{$rr->CompStateChange}){
    (my $p, my $sR, my $sP)= @{$sc};
    $p= $map_reverse{$p};
    $ostring.= $indent3."<StateChange";
    $ostring.= " site=\"".pointer_to_ID($id."_R",$p)."\"";
    $ostring.= " finalState=\"".$sP."\"";
    $ostring.= "/>\n";
  }

  # Binding
  for my $ea (@{$rr->EdgeAdd}){
    (my $p1, my $p2)= @{$ea};
    $p1= $map_reverse{$p1};
    $p2= $map_reverse{$p2};
    $ostring.= $indent3."<AddBond";
    $ostring.= " site1=\"".pointer_to_ID($id."_R",$p1)."\"";
    $ostring.= " site2=\"".pointer_to_ID($id."_R",$p2)."\"";
    $ostring.= "/>\n";
  }

  # Dissociation
  for my $ed (@{$rr->EdgeDel}){
    (my $p1, my $p2)= @{$ed};
    $ostring.= $indent3."<DeleteBond";
    $ostring.= " site1=\"".pointer_to_ID($id."_R",$p1)."\"";
    $ostring.= " site2=\"".pointer_to_ID($id."_R",$p2)."\"";
    $ostring.= "/>\n";
  }

  # Molecule Deletion
  for my $molref (@{$rr->MolDel}){
    $ostring.= $indent3."<Delete";
    $ostring.= " id=\"".pointer_to_ID($id."_R",$molref)."\"";
    $ostring.= "/>\n";
  }

  # Species Creation
  for my $molAddRef (@{$rr->MolAdd}){
    $ostring.= $indent3."<Add";
    my @compInhRefs= @$molAddRef;
    my $molRef= shift(@compInhRefs);
    $ostring.= " id=\"".pointer_to_ID($id."_P",$molref)."\"";
    $ostring.= "/>\n";
  }

  $ostring.=$indent2."</ListOfOperations>\n";

  # Termination
  if ($ostring){
    $string.=">\n"; # terminate tag opening
    $string.= $ostring;
    $string.=$indent."</ReactionRule>\n";
  } else {
    $string.="/>\n"; # short tag termination
  }
}

sub pointer_to_ID{
  my $id=shift; # Use first string as header
  my $pointer= shift;
  my @names= ("P","M","C");

  my @arr= split('\.',$pointer);
  if ($arr[0]==-1){
    $id="Null";
  } else {
    $id.= sprintf "%s%d", $names[0],$arr[0]+1;
    for my $i (1..$#arr){
      $id.= sprintf "_%s%d", $names[$i],$arr[$i]+1;
    }
  }
  return($id);
}

sub findMap{
  my $rr= shift;
  my $mtlist= shift;

  # Aggregate reactant and product graphs
  my $rg= SpeciesGraph::copy(@{$rr->Reactants});
  my $pg= SpeciesGraph::copy(@{$rr->Products});
  #printf "%s -> %s\n", $rg->toString(), $pg->toString();

  # Set up mapping between molecules of aggregate graphs and molecules of patterns
  @aggMapR=();
  @aggMapP=();
  *patterns= $rr->Reactants;
  for $ipatt (0..$#patterns){
    *molecules= $patterns[$ipatt]->Molecules;
    for $imol (0..$#molecules){
      push @aggMapR, "$ipatt.$imol";
      #print "$#aggMapR -> $ipatt.$imol\n";
    }
  }
  *patterns= $rr->Products;
  for $ipatt (0..$#patterns){
    *molecules= $patterns[$ipatt]->Molecules;
    for $imol (0..$#molecules){
      push @aggMapP, "$ipatt.$imol";
      #print "$#aggMapP -> $ipatt.$imol\n";
    }
  }

  # Find map between reactant and product graphs
   my @maps= $rg->findMaps2($pg);
   if (!@maps){
     return("No valid mapping could be found for this rule.");
   }
   if (scalar(@maps)>1){
     return("Mapping for this rule is ambiguous.");
   }

  # Set scaling of reaction multiplicity for rule based on symmetry of reactant
  # matching patterns that is preserved in the mapping onto product patterns

  # Find reactant graph automorphisms
  my @r_auto= $rg->isomorphicToSubgraph($rg);
  my $n_auto_r= scalar(@r_auto);

  # Find product graph automorphisms
  my @p_auto= $pg->isomorphicToSubgraph($pg);
  my $n_auto_p= scalar(@p_auto);

  #printf "$n_auto_r $n_auto_p\n";

  # Set up product autmorphism hash (includes identity- trivial autmorphism)
  my %p_auto_hash;
  %p_auto_hash= ();
  for my $imap (0..$#p_auto){
    my $map= $p_auto[$imap];
    #printf "Product automorphism %d\n" , $imap;
    #print $map->toString(),"\n";
    $p_auto_hash{$map->toString()}=1;
  }

  # Count number of reactant automorphisms that map to product side.  The
  # identity always maps.
  my $n_auto_r_map=1;
  for my $imap (1..$#r_auto){
    my $map= $r_auto[$imap];
    #printf "Reactant automorphism %d\n" , $imap;
    #print $map->toString(),"\n";
    # Transform automorphism
    $map->mapAuto($maps[0]);
    #printf "Transformed automorphism %d\n" , $imap;
    #print $map->toString(),"\n";
    if ($p_auto_hash{$map->toString()}){
      ++$n_auto_r_map;
    }
  }
  my $multScale= 1/$n_auto_r_map;

  # OBSOLETE:
  # This formula assumes that all automorphisms in found in the smaller
  # automorphism group have corresponding automorphisms in the larger group.
  # That is, if there are 2 autmorphisms on the reactant side and 2 on the
  # product side, then each of the reactant automorphisms has a corresponding
  # product autmorphism.
  #my $multScale= 1/MIN($n_auto_r,$n_auto_p);

  $rr->MultScale($multScale);
  #printf "Multscale= $multScale\n";

  *moleculesR= $rg->Molecules;
  *moleculesP= $pg->Molecules;
  *map= $maps[0]->MapF;

  # Use mapping to determine graph rewriting steps
  #printf "Found %d mapping(s)\n", scalar(@maps);
#  printf $maps[0]->toString();

  # Molecules that are destroyed
  my @dlist, @pused;
  @dlist=();
  @pused=(0)x@moleculesP;

  *patterns= $rr->Reactants;
  my $i_r=0;
  for my $p_r (0..$#patterns){
    *molecules= $patterns[$p_r]->Molecules;
    my @mol_delete;
    @mol_delete=();
    for my $m_r (0..$#molecules){
      my $i_p= $map{$i_r};
      if ($i_p<0){
	push @mol_delete, $m_r;
      } else {
	$pused[$i_p]=1;
      }
      ++$i_r;
    }
    if ($rr->DeleteMolecules || ($#mol_delete<$#molecules)){
      # Delete only molecules by placing pointer to each on deletion list
      for $m_r (@mol_delete){
	push @dlist, "${p_r}.${m_r}";
 	#printf "Rule %s: Reactant molecule %s is deleted\n", $rr->Name, $rr->Reactants->[$p_r]->Molecules->[$m_r]->toString();
      }
    } else {
      # Delete whole species matching pattern
      #printf "Rule %s: Reactant pattern %s is deleted\n", $rr->Name, $rr->Reactants->[$p_r]->toString();
      push @dlist, $p_r;
    }
  }
  $rr->MolDel([@dlist]);


  # Molecules that are created
  @molAdd=();
  *mapr= $maps[0]->MapR;
  for $imP (0..$#pused){
    next unless ($pused[$imP]==0);
    # Check that this molecule is fully specified, i.e. all components and component states
    my $ref= $aggMapP[$imP];
    my ($p,$m)= split('\.',$ref);
    my $mol= $rr->Products->[$p]->Molecules->[$m];
    #printf "Rule %s: Product molecule %s is created\n", $rr->Name,$mol->toString();
    # List of pointers to reactant components from which to inherit attributes of each component
    my $ic=0;
    my @ilist=();
    for my $comp (@{$mol->Components}){
      if (($cref=$mapr{"$imP.$ic"}) ne "-1"){
	#print "cref=$cref\n";
	my ($imR, $icR)= split('\.',$cref);
	$cref= $aggMapR[$imR].".$icR";
	push @ilist, $cref;
	#printf "   Component $ic inherits from reactant pointer $cref\n";
      } else {
	# Is component state specified
	push @ilist, "";
      }
      ++$ic;
    }
    if ($err=$mtlist->checkMolecule($mol,{IsSpecies=>1,InheritList=>\@ilist})){
      return("Molecule created in reaction rule: ".$err);
    }
    push @molAdd, [($ref,@ilist)];
  }
  if (@molAdd){
    $rr->MolAdd([@molAdd]);
  }

  # Edges
  *edgesR= $rg->Edges;
  *adjR= $rg->Adjacency;
  *edgesP= $pg->Edges;
  *adjP= $pg->Adjacency;

  #printf "%d edges reac %d edges prod\n", scalar(@edgesR), scalar(@edgesP);

  # Edges deleted
  @edgeDel=();
  @eused=(0)x@edgesP;
  for $edgeR (@edgesR){
    my @ps= split(' ',$edgeR);
    next unless ($#ps==1);
    my $p1R= $ps[0];
    my $p2R= $ps[1];
    $p1P= $map{$p1R}; 
    $p2P= $map{$p2R}; 
    if (defined($adjP{$p1P}{$p2P})){
      $eused[$adjP{$p1P}{$p2P}]=1;
      next;
    }
    ($im1, $ic1)= split('\.',$p1R);
    ($im2, $ic2)= split('\.',$p2R);
    $p1R= $aggMapR[$im1].".$ic1";
    $p2R= $aggMapR[$im2].".$ic2";
    #printf "Edge $p1R $p2R is deleted (reactant side)\n";
    push @edgeDel, [$p1R, $p2R];
  }
  if (@edgeDel){
    $rr->EdgeDel([@edgeDel]);
  }

  # Edges added : Need to convert pointers to products back to pointers to reactants
  # There is a problem for molecules that appear in products but not in reactants.
  # For these, the only way to add the edge is to assign the mapping of product molecules 
  # onto the selected graphs
  @edgeAdd=();
  for $iedgeP (0..$#edgesP){
    next unless ($eused[$iedgeP]==0);
    my @ps= split(' ',$edgesP[$iedgeP]);
    next unless ($#ps==1);
    my $p1P= $ps[0];
    my $p2P= $ps[1];
    ($im1, $ic1)= split('\.',$p1P);
    ($im2, $ic2)= split('\.',$p2P);
    $p1P= $aggMapP[$im1].".$ic1";
    $p2P= $aggMapP[$im2].".$ic2";
    #printf "Rule %s: Edge %s %s is added (product side)\n", $rr->Name,$p1P, $p2P;
    push @edgeAdd, [$p1P, $p2P];
  }
  if (@edgeAdd){
    $rr->EdgeAdd([@edgeAdd]);
  }

  # Component state changes
  @compStateChange=();
  for $imR (0..$#moleculesR){
    $imP=$map{$imR};
    #print "$imR->$map{$imR}\n";
    next unless ($imP>=0);
    *componentsR= $moleculesR[$imR]->Components;
    *componentsP= $moleculesP[$imP]->Components;
    for $icR (0..$#componentsR){
      next unless (defined($componentsR[$icR]->State));
      $stateR=$componentsR[$icR]->State;
      ($imP,$icP)= split('\.',$map{"$imR.$icR"});
      if (!defined($componentsP[$icP]->State)){
	exit_error("Component with state attribute defined in reactant pattern cannot map to component
 with undefined state attribute in product pattern", $rr->toString());
      } else {
	$stateP=$componentsP[$icP]->State;
      }

      if ($stateR cmp $stateP){
	$pR= "$aggMapR[$imR].$icR";
	#printf "Component %s changed state from %s to %s\n", $pR, $stateR, $stateP;
	push @compStateChange, [($pR, $stateR, $stateP)];
      }
    }
  }
  if (@compStateChange){
    $rr->CompStateChange([@compStateChange]);
  }

  # Compartment changes
  my @changeCompartment=();
  

  if (@changeCompartment){
    $rr->ChangeCompartment([@changeCompartment]);
  }

  # Rewrite map to entries into form
  # map{pR.mR.cR}= "pP.mP.cP"; 
  my %map_new=();
  for $im1 (0..$#moleculesR){
    $im2= $map{$im1};
    $pm1= $aggMapR[$im1];
    $pm2= ($im2>=0) ? $aggMapP[$im2] : $im2;
    $map_new{$pm1}= $pm2;
    #print "map_new: $pm1 -> $pm2\n";
    *components= $moleculesR[$im1]->Components;
    for $ic1 (0..$#components){
      ($im2, $ic2)= split('\.',$map{"$im1.$ic1"});
      if ($im2>=0){
	$map_new{"$pm1.$ic1"}= "$pm2.$ic2";
	#print "map_new: $pm1.$ic1 -> $pm2.$ic2\n";
      } else {
	$map_new{"$pm1.$ic1"}= $im2;
	#print "map_new: $pm1.$ic1 -> $pm2.$ic2\n";
      }
    }
  }
  $rr->Map({%map_new});

  return("");
}

{
  my $pattDepth=-1, $maxDepth;
  my $check_iso=1, $max_agg=1e99, $max_stoich="", $verbose;
  my $rr, @sgs, @matches, @isnew, $new_level, @targets, @reactants, @offsets=(0);
  my $n_new_rxns;
  my %mapP;
  my $slist, $rlist; # RxnList and SpeciesList
  my $rtype, $dlist;

  # $rule->applyRule(SpeciesList, RxnList, (Species))
  sub applyRule{
    # Only need to read arguments when $pattDepth is negative
    if ($pattDepth==-1){
      $rr=shift;
      $slist= shift;
      $rlist= shift;
      $maxDepth=$#{$rr->Reactants};
      $n_new_rxns=0;

      # Handle deletion rules
      $rtype= $rr->RateLaw->Type;
      if ($rtype eq "Zero"){
	$dlist= RxnList->new;
      } else {
	$dlist="";
      }

      # Determine to which species rules are to be applied
      if (@_){ # User-provided array
	*species=shift;
      } else { # Array in species list
	*species= $slist->Array;
      }
      @sgs=();
      for $spec (@species){
	if (!($spec->RulesApplied)){
	  push @sgs, $spec->SpeciesGraph;
	}
      }
      
      # Next argument is optional parameter hash
      # Any parameters that don't have defaults set here will inherit values set 
      # by previous calls to this function
      # SAFER would be to make these attributes of BNGModel and automatically
      # pass them to this function if they are defined
      if (@_){
	my $params=shift;
	if (defined($params->{max_agg})){
	  $max_agg= $params->{max_agg};
	} else {
	  #$max_agg=1e99;
	}
	if (defined($params->{check_iso})){
	  $check_iso= $params->{check_iso};
	} else {
	  #$check_iso=1;
	}
	if (defined($params->{max_stoich})){
	  $max_stoich=$params->{max_stoich};
	} else {
	  #$max_stoich="";
	}
	if (defined($params->{verbose})){
	  $verbose= $params->{verbose};
	} else {
	  #$verbose=0;
	}
      }

      @rxns=();
      %mapP=();

      # Map rule patterns onto SpeciesGraphs
      my $ipatt=0;
      *rmatches= $rr->Rmatches;
      my $new_matches_found=0;
      $new_level=-1;
      my $nsets=1;
      for my $patt (@{$rr->Reactants}){
	my @matches=();
	my $include_default= (@{$rr->Rinclude->[$ipatt]}) ? 0 : 1;
	my $targ_last="";
	my $include_last="";
	# Check if $match->Target has subgraph isomorphic to excluded graph
	for my $match ($patt->isomorphicToSubgraph(@sgs)){
	  my $targ= $match->Target;
	  my $include=$include_default;
	  # Only do new isomorphism check if $targ has changed
	  if ($targ!=$targ_last){ 
	    for my $patt_incl (@{$rr->Rinclude->[$ipatt]}){
	      if ($patt_incl->isomorphicToSubgraph($targ)){
		# Include this match from list since graph matches included graph
		#print "Reactant ", $targ->toString()," matches included pattern\n";
		$include=1;
		last;
	      }
	    }
	    for my $patt_excl (@{$rr->Rexclude->[$ipatt]}){
	      if ($patt_excl->isomorphicToSubgraph($targ)){
		# Exclude this match from list since graph matches excluded graph
		#print "Reactant ", $targ->toString()," matches excluded pattern\n";
		$include=0;
		last;
	      }
	    }
	  } else {
	    $include=$include_last;
	  }
	  if ($include){
	    push @matches, $match;
	    #print "Included ", $targ->toString(),"\n";
	  } else {
	    #print "Excluded ", $targ->toString(),"\n";
	  }
	  $targ_last= $targ;
	  $include_last= $include;
	}
	my $n_new= scalar(@matches);
	$new_matches_found+= $n_new;
	if ($n_new){
	  $new_level= $ipatt;
	}
	push @{$rmatches[$ipatt]}, @matches;
	my $n_tot= scalar(@{$rmatches[$ipatt]});
	$nsets*= $n_tot;
	if ($verbose){printf " %4d new (%4d total) matches for reactant %d\n", $n_new, $n_tot, $ipatt+1;}
	++$ipatt;
      }
      #print "new_level=$new_level max_level= $maxDepth\n";

      # Don't do any more work unless $matches_found
      if (!$new_matches_found){
	#print "No new matches found.\n";
	return(0);
      }
      # Don't do any more work unless at least one possible reactant set
      # ($nsets) is found
      if (!$nsets){
	#print "No complete match sets.\n";
	return(0);
      }

      ++$pattDepth;
      @matches=();
      @isnew=();
      @targets=();
      @reactants=();
    }

    # Loop over matches at current level
  MATCH:
    for $match (@{$rmatches[$pattDepth]}){
      $matches[$pattDepth]= $match;
      $targets[$pattDepth]= $match->Target;
      $reactants[$pattDepth]= $match->Target->Species;
      $isnew[$pattDepth]= !($match->Target->Species->RulesApplied);
      #printf "isnew=%d\n", $match->Target->Species->RulesApplied;
      $offset= $offsets[$pattDepth];

      # Only use current set if it contains a match to a new species or there are new
      # matches below the current level
      my $new_set=0;
      for my $depth (0..$pattDepth){
	if ($isnew[$depth]){
	  $new_set=1;
	  last;
	}
      }
      # Skip set if no new matches below the current level
      if (!$new_set){
	if ($new_level<=$pattDepth){
	  #print "skip\n";
	  next MATCH;
	}
      }
      
      # Set mapping of product pattern onto supergraph
      # Loop over molecules of pattern
      # Loop over components of molecule
      *molecules= $match->Source->Molecules;
      *map_RtoP= $rr->Map;
      #print "Mapping ", $match->Source->toString(), " onto ", $match->Target->toString(),"\n";
      for my $m (0..$#molecules){
	*components= $molecules[$m]->Components;
	if (defined($map_RtoP{"$pattDepth.$m"})){
	  $pP= $map_RtoP{"$pattDepth.$m"};
	  next if ($pP==-1);
	} else {
	  # skip this molecule pattern if it doesn't map into products
	  #print "gothere\n";
	  next;
	}
	my $targ= $match->mapF("$m", $offset);
	$mapP{$pP}= $targ;
	#print "$pP -> $mapP{$pP}\n";
	for my $c (0..$#components){
	  # Get pointer to corresponding component of product pattern
	  my $pP= $map_RtoP{"$pattDepth.$m.$c"};
	  # Set map of product pointer with pointer in Target 
	  $mapP{$pP}= $match->mapF("$m.$c", $offset);
	  #print "$pattDepth.$m.$c -> $pP -> $mapP{$pP}\n"
	}
      }

      # Descend to next level
      if ($pattDepth<$maxDepth){
	++$pattDepth;
	$offsets[$pattDepth]= $offset + scalar(@{$match->Target->Molecules});;
	&applyRule;
	next;
      }

      # Apply rule transformations using current set of maps

      # Aggregate matching SpeciesGraphs
      my $g= SpeciesGraph::copy(@targets);
      *mol= $g->Molecules;

      # Apply component state changes
      for $sc (@{$rr->CompStateChange}){
	$ref= $sc->[0];
	#$stateR= $sc->[1];
	$stateP= $sc->[2];
	my ($p, $m, $c)= split('\.', $ref);
	($im2, $ic2)= split('\.', $matches[$p]->mapF("$m.$c", $offsets[$p]));
	$mol[$im2]->Components->[$ic2]->State($stateP);
      }

      # Apply edge deletions
      for my $edel (@{$rr->EdgeDel}){
	my @refs= @$edel;
	my @ps=();
	for $ref (@refs){
	  my ($p,$m,$c)= split('\.', $ref);
	  # Find pointer to molecule and component in $g corresponding to
	  # $m.$c in pattern
	  push @ps, $matches[$p]->mapF("$m.$c", $offsets[$p])
	}
	#print "Deleting edge $ps[0] $ps[1]\n";
	$g->deleteEdge(@ps);
      }

      # Apply molecule additions
      for my $molAddRef (@{$rr->MolAdd}){
	my @compInhRefs= @$molAddRef;
	my $molRef= shift(@compInhRefs);
	my ($p,$m)= split('\.',$molRef);
	my $newMol= $rr->Products->[$p]->Molecules->[$m]->copy($molRef);

	# Inherit component atributes
	my $ic=0;
	for my $compInhRef (@compInhRefs){
	  next unless $compInhRef;
	  my ($p,$m,$c)= split('\.', $compInhRef);
	  ($im2, $ic2)= split('\.', $matches[$p]->mapF("$m.$c", $offsets[$p]));
	  # Copy component state
	  my $stateNew= $mol[$im2]->Components->[$ic2]->State;
	  $newMol->Components->[$ic]->State( $stateNew);
	  ++$ic;
	}

	# Remove labels and set map elements
	$newMol->Label(undef);
	my $nmol= $#{$g->Molecules}+1;
	$mapP{$molRef}= $nmol;
	my $ic=0;
	for my $comp (@{$newMol->Components}){
	  $comp->Label(undef);
	  $mapP{"$molRef.$ic"}= "$nmol.$ic";
	  ++$ic;
	}

	push @{$g->Molecules}, $newMol;
	# printf "prod: %s\n", $g->toString();
	# Should add pointer from product pattern to new molecules in $g
      }


      # Apply edge additions
      my $nedge=1;
      for my $eadd (@{$rr->EdgeAdd}){
	my @refs= @$eadd;
	$p1= $mapP{$refs[0]};
	$p2= $mapP{$refs[1]};
	#printf "Adding edge p1=$p1(%s) p2=$p2(%s)\n", $refs[0],$refs[1];
	$g->addEdge("ne${nedge}", $p1, $p2);
	++$nedge;
      }

      # Apply molecule/pattern deletions
      my @del_list=();
      for my $molref (@{$rr->MolDel}){
	my ($p,$m)= split('\.',$molref);
	if (defined($m)){
	  # Delete molecule
	  my $mi= $matches[$p]->mapF("$m", $offsets[$p]);
	  #print "Deleting molecule ", $g->Molecules->[$mi]->toString(),"\n";
	  $g->Molecules->[$mi]="";
	  push @del_list, $mi;
	} else {
	  # Delete delete whole pattern
	  $mi_max= $offsets[$p]+$#{$matches[$p]->Target->Molecules};
	  for my $mi ($offsets[$p]..$mi_max){
	    #print "Deleting molecule ", $g->Molecules->[$mi]->toString(),"\n";
	    $g->Molecules->[$mi]="";
	    push @del_list, $mi;
	  }
	}
      }
      my $imol=0;
      *molecules= $g->Molecules;
      while ($imol<=$#molecules){
	if ($molecules[$imol] eq ""){
	  splice @molecules,$imol,1;
	  next;
	}
	++$imol;
      }

      # Update aggregate graph (and remove dangling edges)
      $g->updateEdges(0);

      # Split aggregate graph into connected subgraphs
      my ($n_subgraphs, $subgraphs)= $g->findConnected2();

      # Check for correct number of subgraphs
      my $nprod_patterns= scalar(@{$rr->Products});
      if ($n_subgraphs != $nprod_patterns){
	# Having more subgraphs than product graphs is allowed if Molecules
	# are being deleted.
	if (!($rr->DeleteMolecules) && ($n_subgraphs < $nprod_patterns)){
	  printf "Rule %s: n_sub ($n_subgraphs)!= n_prod ($nprod_patterns): %s\n", $rr->Name, $g->toString();
	  next MATCH;
	} else {
	  #printf "Deleting molecules: %s\n", $g->toString();
	}
      }

      # map of target molecules in supergraph back onto product patterns
      # useful for checking below
      @mapP_reverse=(0)x@$subgraphs;
      for my $p (keys %mapP){
	my $targ= $mapP{$p};
	next if $targ=~/\./;
	$mapP_reverse[$targ]=$p;
      }
      # Correct $mapP_reverse for deleted molecules
      for my $mi (@del_list){
	$mapP_reverse[$targ]="d";
      }
      my $it=0;
      while ($it<=$#mapP_reverse){
	if ($mapP_reverse[$it] eq "d"){
	  splice @mapP_reverse,$it,1;
	  next;
	}
	++$it;
      }
      #for my $it (0..$#mapP_reverse){
      #	printf "it=$it p=%s\n", $mapP_reverse[$it];
      #}


      # Loop over product graphs
      my @products=();
      my $new_prod=0;
      my $iprod;
  PRODUCT:
      for my $isub (1..$n_subgraphs){
	# To what product pattern does this subgraph correspond?
	# Check that molecules from only a single product pattern map onto
	# this subgraph, i.e. that molecules of graph partition correctly into products
	my $ip,$ip_sav,$im;
	$ip_sav=-1;
	$ip=-1;
	for my $imol (0..$#{$subgraphs}){
	  next unless ($subgraphs->[$imol]==$isub);
	  $p= $mapP_reverse[$imol];
	  #print "p=$p\n";
	  next unless ($p ne "0");
	  ($ip,$im)= split('\.',$p);
	  #printf "isub=$isub imol=$imol ip=$ip ip_sav=$ip_sav pP=$p %s\n", $mapP{$p};
	  if ($ip_sav==-1){
	    $ip_sav=$ip;
	    # Skip this product if it has Fixed attribute
	    # This avoids change of concentration
	    if ($rr->Products->[$ip]->Fixed){
	      #printf "isub=$isub imol=$imol ip=$ip ip_sav=$ip_sav pP=$p %s nsub=%d\n", $mapP{$p},$n_subgraphs;
	      next PRODUCT;
	    }
	  }
	  elsif ($ip != $ip_sav){
	    # Put informative warning message here
	    #printf "isub=$isub imol=$imol ip=$ip ip_sav=$ip_sav pP=$p %s imax_mol=%d\n", $mapP{$p},$#{$subgraphs};
	    # print $g->toString(),"\n";
	    next MATCH;
	  }
	}
	$iprod=$ip;

	# This might be unnecessary and inefficient
	my $p= ($n_subgraphs>1) ? $g->copySubgraph($subgraphs,$isub) : $g;

	if ($#{$p->Molecules}>=$max_agg){ 
	  #printf "%d molecules exceeds max_agg of %d\n", $#{$p->Molecules}+1, $max_agg;
	  next MATCH;
	}
	if ($max_stoich){
	  for my $key (keys %{$max_stoich}){
	    my $max= $max_stoich->{$key};
	    next if $max=~/^unlimited/;
	    if ($p->stoich($key) > $max){
	      #printf "Stoichometry of $key in %s exceeds max of %d\n", $p->toString(), $max;
	      next MATCH;
	    }
	  }
	}

	# Put product graph in canonical order (quasi-canonical for the time being)
	if (my $err=$p->sortLabel()){
	  # Only here to debug Compartments; should never be reached.
	  die "$err";
	}
	
	# Does product match excluded pattern?
	if ($iprod>=0){
	  for my $patt_excl (@{$rr->Pexclude->[$iprod]}){
	    if ($patt_excl->isomorphicToSubgraph($p)){
	      # Abort this reaction 
	      #print "Product ", $p->toString()," matches excluded pattern\n";
	      next MATCH;
	    }
	  }

	  # Does product match included pattern?  Must do so if include patterns are specified.
	  my $include=1;
	  for my $patt_incl (@{$rr->Pinclude->[$iprod]}){
	    if ($patt_incl->isomorphicToSubgraph($p)){
	      $include=1;
	      last;
	    } else {
	      $include=0;
	    }
	  }
	  next MATCH unless $include;
	}

	# Update species list; add checks whether this sg isomorphic to
	# existing species and creates new species if not.  It returns
	# a pointer to new or existing species with isomorphic sg.
	# If we want mapping for this reaction need to update to
	# account for reordering of molecules and components
	my $spec;
	if (!($spec=$slist->lookup($p, $check_iso))){
	  $spec= $slist->add($p, 0);
	  ++$new_prod;
	}
	if ($iprod>=0){
	  if (defined($products[$iprod])){
	    push @products, $products[$iprod];
	  }
	  $products[$iprod]= $spec;
	} else {
	  push @products, $spec;
	}
	#for my $spec (@products){
	#  print " ",$spec->SpeciesGraph->toString();
	#}
	#print "\n";
      }

      # Add any reactants with Fixed attribute to products list
      # to insure concentration does not change
      my $ri=0;
      for my $rpatt (@{$rr->Reactants}){
	if($rpatt->Fixed){
	  #print "Fixed reactant ",$rpatt->toString(),"\n";
	  push @products, $reactants[$ri];
	}
	++$ri;
      }

      # Add reaction to rlist 
      my $rxn= Rxn->new();
      $rxn->Reactants([@reactants]);
      $rxn->Products([@products]);
      $rxn->RateLaw($rr->RateLaw);
      $rxn->StatFactor($rr->MultScale);
      $rxn->Priority($rr->Priority);
      $rxn->RxnRule($rr);
      if ($dlist){
	$dlist->add($rxn,1);
      } else {
	$n_new_rxns+= $rlist->add($rxn);
	if ($n_new_rxns==0){print "Invalid reaction found: ", $rxn->toString(),"\n";}
      }

#        # Print reaction
#        print "new reaction: ";
#        my $itarg=0;
#        for my $targ (@targets){
#  	print " + " unless $itarg==0;
#  	print $targ->toString();
#  	++$itarg;
#       }
#        print " -> ";
#        my $iprod=0;
#        for my $prod (@prods){
#  	print " + " unless $iprod==0;
#  	print $prod->SpeciesGraph->toString();
#  	++$iprod;
#        }
#        print "\n";

#      for $i (0..$pattDepth){
#	print " $i $offsets[$i] ",$matches[$i]->Target->toString();
#      }

    }

    # Ascend if not at top
    if ($pattDepth>0){
      --$pattDepth;
      return;
    }

    # Correct multiplicity of generated reactions based on sizes of
    # automorphism groups for reactant and product pattern graphs 

    # Delete reactions if this rule has RateLaw->Type= Zero
    if ($dlist){
      for my $rxn (@{$dlist->Array}){
	$n_new_rxns+= $rlist->add($rxn);
      }
    }

    $pattDepth=-1;
    return($n_new_rxns);
  }
}

1;
