#!/usr/bin/perl
# Validation script for BioNetGen suite.
# Type "perl validate_examples.pl" in the Validate directory to execute.

# TODO:
# 1) Add SBML validator.

$exec="../Perl2/BNG2.pl";
#$exec="../PerlBin/BNG2";
$verify="../Perl2/verify.pl";
$logfile= "validation_log";
$datdir= "DAT_validate";

$skip_run=0;

# turn off output buffering on STDOUT
(select(*STDOUT), $|=1)[0];

# Process command line arguments
while ($ARGV[0] =~ /^-/){
  $_ = shift;
  if (/^-exec$/){
    $exec= shift;
  }
  elsif (/^-skip$/){
    $skip_run=1;
  }
  else{
    exit_error("Unrecognized command line option $_");
  }
}

# set up logfile
open(h,">$logfile");
close(h);

# Handling of reaction multiplicities
print "Handling of reaction multiplicities...";
system("$exec tlmr.bngl >> $logfile");
$result= `$verify tlmr.cdat $datdir/tlmr_spec.cdat`;
print "$result";

# Handling of parameter names starting with digit
print "Handling of parameter names...";
system("$exec test_paramname.bngl >> $logfile");
$result= `$verify test_paramname.cdat $datdir/test_paramname.cdat`;
print "$result";

# Correct ordering of reactants with MM or Sat RateLaw
print "Correct ordering of reactants in MM RateLaw reactions...";
system("$exec test_MM.bngl >> $logfile");
$result= `$verify test_MM.cdat $datdir/test_MM.cdat`;
print "$result";

print "Correct ordering of reactants in Sat RateLaw reactions...";
system("$exec test_Sat.bngl >> $logfile");
$result= `$verify test_Sat.cdat $datdir/test_Sat.cdat`;
print "$result";

print "Correct reading of previously written net file...";
system("$exec test_read_sat.bngl >> $logfile");
$result= `$verify test_read_sat.cdat $datdir/test_read_sat.cdat`;
print "$result";

print "Correct handling of reactions using DeleteMolecules...";
my $mess="PASSED";
system("$exec deleteMolecules.bngl >> $logfile");
if (system("./compare_species.pl deleteMolecules.net $datdir/deleteMolecules.net >> $logfile")){
  $mess="FAILED";
}
if (system("./compare_rxn.pl deleteMolecules.net $datdir/deleteMolecules.net >> $logfile")){
  $mess="FAILED";
}
print $mess,"\n";


# Test specific models
@models=qw(
toy-jim
Haugh2b
egfr_net
fceri_ji
SHP2_base_model
);

#$skip_run=1;

for $model (@models){
  print "Testing model $model\n";
  $skip_run || system("$exec $model.bngl > $model.log");
  print "Observables...";
  $result= `$verify $model.gdat $datdir/$model.gdat`;
  print "$result";
  print "All concentrations...";
  $result= `$verify $model.cdat $datdir/$model.cdat`;
  print "$result";
}
